/********************************************************************************
 *   dbgreader -- system debug output reader for Windows CE                     *
 *                                                                              * 
 *   Copyright (C) 2004-2007 dci (dev@cpp.in)                                   *
 *                                                                              *
 *   This program is free software; you can redistribute it and/or              *
 *   modify it under the terms of the GNU General Public License version 2      *
 *   as published by the Free Software Foundation.                              *
 *                                                                              *
 *   This program is distributed in the hope that it will be useful,            *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU          *
 *   General Public License version 2 for more details.                         *
 *                                                                              *
 *   You should have received a copy of the GNU General Public License          *
 *   version 2 along with this library; if not, write to the Free Software      *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  *
 ********************************************************************************
 */


/*
 You may wonder why kernel sync objects aren't used here? Well, they were,
 but it seems that WriteDebugStringFunc is sometimes called with disabled
 interrupts, so any syscall hangs the device.
 Maybe a kernel thread should be created which will be in this weird sync
 with WriteDebugStringFunc and will pulse some control event when my sync
 is fired, but for now it's implemented in this way.
*/

#include <windows.h>

/* shared source headers */
#include <pkfuncs.h>
#include <psyscall.h>
#include <kernel.h>

/* */
#include "rt.h"
#include "staticpool.h"

#include "dbgreader.h"


/* msg queue */
typedef struct qhdr_tag
{
	unsigned short msg_id;
	unsigned short msg_refs;
}
qhdr_t;

qhdr_t * g_msg_queue [ 512 ];
unsigned short g_msg_last;


void msg_push ( wchar_t * msg )
{
	/* create block */
	qhdr_t * block = dsp_alloc ( sizeof ( qhdr_t ) + sizeof ( msginfo_t ) + ( ( mywcslen ( msg ) + 1 ) * sizeof ( * msg ) ) );
	if ( block )
	{
		unsigned int i, numitems = sizeof ( g_msg_queue ) / sizeof ( * g_msg_queue );
		qhdr_t * oldmsg = NULL;
		unsigned long procmode;
		/* fill */
		msginfo_t * info = ( msginfo_t * ) ( block + 1 );
		block->msg_id = ++g_msg_last;
		block->msg_refs = 1; /* refd by queue */
		/* */
		procmode = cpsr_get ( ) & SR_MODE_MASK;
		if ( ( FIQ_MODE == procmode ) || ( IRQ_MODE == procmode ) || ( SVC_MODE == procmode ) )
		{
			info->dwProcessId = info->dwThreadId = 0;
		}
		else
		{


			/* seems binary compatible for ce4<->ce5 */
			info->dwProcessId = ( unsigned long ) KData.ahSys [ SH_CURPROC ];
			info->dwThreadId  = ( unsigned long ) KData.ahSys [ SH_CURTHREAD ];


		}
		mywcscpy ( ( wchar_t * ) ( info + 1 ), msg );
		/* try to push to the end */
		for ( i = 0; i < numitems; ++i )
		{
			if ( NULL == g_msg_queue [ i ] )
			{
				g_msg_queue [ i ] = block;
				return;
			}
		}
		/* no success. need to kick out some old stuff */
		oldmsg = * g_msg_queue;
		for ( i = 1; i < numitems; ++i ) g_msg_queue [ i - 1 ] = g_msg_queue [ i ];
		g_msg_queue [ numitems - 1 ] = block;
		if ( 0 == --oldmsg->msg_refs ) dsp_free ( oldmsg );
	}
}

unsigned short msg_last ( )
{
	return g_msg_last;
}

size_t msg_read ( unsigned short msgid, msginfo_t * buf, size_t bufsize )
{
	unsigned int i, numitems = sizeof ( g_msg_queue ) / sizeof ( * g_msg_queue );
	if ( bufsize < ( sizeof ( msginfo_t ) + sizeof ( wchar_t ) ) ) return 0;
	for ( i = 0; i < numitems; ++i )
	{
		qhdr_t * block = g_msg_queue [ i ];
		if ( msgid == block->msg_id )
		{
			size_t ret;
			msginfo_t * info = ( msginfo_t * ) ( block + 1 );
			++block->msg_refs;
			buf->dwProcessId = info->dwProcessId;
			buf->dwThreadId  = info->dwThreadId;
			ret = mywcsncpy ( ( wchar_t * ) ( buf + 1 ), ( wchar_t * ) ( info + 1 ), ( bufsize - sizeof ( msginfo_t ) ) / sizeof ( wchar_t ) );
			if ( 0 == --block->msg_refs ) dsp_free ( block );
			return sizeof ( msginfo_t ) + ( ( ret + 1 ) * sizeof ( wchar_t ) );
		}
	}
	return 0;
}


/* hooking ctl */
long dwState = HS_DISABLE;

DWORD pProcPtr = 0;
DWORD pOrigProc = 0;
DWORD dwHookingTest = 0;

DWORD hooking_test ( )
{
	return dwHookingTest;
}

long hooking_get_state ( )
{
	return dwState;
}

long hooking_set_state ( long newstate )
{
	long ret = dwState;
	dwState = newstate;
	if ( pProcPtr ) * ( ( DWORD * ) pProcPtr ) = ( HS_DISABLE != newstate ) ? ( DWORD ) msg_push : pOrigProc;
	return ret;
}

void hooking_init ( )
{
	DWORD dwPptr = 0; /* traced addr */
	DWORD dwSig = 0; /* ptr to ptr "(NULL)" */
	DWORD dwSrc = 0; /* tracer start addr */
	DWORD dwKMemStart = 0, dwKMemEnd = 0;
	DWORD * dwPtr; /* tracer */


	/* seems binary compatible for ce4<->ce5
	   tracer will go up from this ptr searching for lpWriteDebugStringFunc in OutputDebugString proc */
	dwSrc = ( DWORD ) ( ( CINFO ** ) KData.aInfo [ KINX_APISETS ] ) [ SH_WIN32 ]->ppfnMethods [ W32_NKvDbgPrintfW ];
	/* seems working for at least WM2003SE, WM5 and WM6. 
	   need something like memoryinfo, but for initialized kernel data here */
	dwKMemStart = 0x80000000;
	dwKMemEnd = 0x89000000;


	/* trace for lpWriteDebugStringFunc */
	for ( dwPtr = ( DWORD * ) dwSrc; dwPtr > ( ( DWORD * ) dwSrc ) - 0x1000; --dwPtr )
	{
		/* signature found and current dword looks like instruction */
		if ( dwSig && ( ( ( * dwPtr ) >> 28 ) == 0xE ) )
		{
			if ( 8 < ( dwSig - ( ( DWORD ) dwPtr ) ) )
			{
				dwPptr = ( DWORD ) ( dwPtr + 1 );
				break;
			}
			else
			{
				dwSig = 0;
			}
		}
		/* current dword is aligned pointer to something in kernel mem */
		if ( ( 0 == ( ( * dwPtr ) & 3 ) ) && ( ( * dwPtr ) >= dwKMemStart ) && ( ( * dwPtr ) <= dwKMemEnd ) )
		{
			/* if it pointer to "(NULL)" ? */
			if ( 0 == mymemicmp ( ( void * ) ( * dwPtr ), L"(null)", 12 ) )
			{
				dwSig = ( DWORD ) dwPtr;
			}
		}
	}
	/* got ptr? */
	if ( dwPptr )
	{
		pProcPtr = * ( DWORD * ) dwPptr;
		pOrigProc = * ( DWORD * ) pProcPtr;
		dwHookingTest = pProcPtr;
	}
}

void hooking_void ( )
{
	hooking_set_state ( HS_DISABLE );
	PRIV_WIN32_CALL( void, Sleep, ( DWORD dwMilliseconds ) ) ( 1000 ); /* let all writers to finish */
}


/* library control routines */
DWORD dwInstLast  = 0;
DWORD dwInstCount = 0;

BOOL IOControl ( DWORD dwInstData, DWORD dwIoControlCode, LPVOID lpInBuf, DWORD nInBufSize, LPVOID lpOutBuf, DWORD nOutBufSize, LPDWORD lpBytesReturned )
{
	if ( IOCTL_DBGSERVER_TEST == dwIoControlCode )
	{
		if ( lpOutBuf && nOutBufSize )
		{
			* ( DWORD * ) lpOutBuf = hooking_test ( );
			if ( lpBytesReturned ) * lpBytesReturned = sizeof ( DWORD );
			return TRUE;
		}
	}
	else if ( IOCTL_DBGSERVER_STATE_GET == dwIoControlCode )
	{
		if ( lpOutBuf && nOutBufSize )
		{
			* ( DWORD * ) lpOutBuf = hooking_get_state ( );
			if ( lpBytesReturned ) * lpBytesReturned = sizeof ( DWORD );
			return TRUE;
		}
	}
	else if ( IOCTL_DBGSERVER_STATE_SET == dwIoControlCode )
	{
		if ( lpInBuf && ( sizeof ( DWORD ) == nInBufSize ) )
		{
			DWORD res = hooking_set_state ( * ( DWORD * ) lpInBuf );
			if ( lpOutBuf && nOutBufSize )
			{
				* ( DWORD * ) lpOutBuf = res;
				if ( lpBytesReturned ) * lpBytesReturned = sizeof ( DWORD );
			}
			return TRUE;
		}
	}
	else if ( IOCTL_DBGSERVER_MSG_LAST == dwIoControlCode )
	{
		if ( lpOutBuf && nOutBufSize )
		{
			* ( unsigned short * ) lpOutBuf = msg_last ( );
			if ( lpBytesReturned ) * lpBytesReturned = sizeof ( unsigned short );
			return TRUE;
		}
	}
	else if ( IOCTL_DBGSERVER_MSG_READ == dwIoControlCode )
	{
		if ( lpInBuf && nInBufSize && lpOutBuf && nOutBufSize )
		{
			size_t res = msg_read ( * ( unsigned short * ) lpInBuf, ( msginfo_t * ) lpOutBuf, nOutBufSize );
			if ( res )
			{
				if ( lpBytesReturned ) * lpBytesReturned = res;
				return TRUE;
			}
		}
	}
	/* */
	return FALSE;
}

DWORD CreateInstance ( )
{
	if ( 0 == dwInstCount++ )
	{
		dsp_init ( );
		hooking_init ( );
	}
	return ++dwInstLast;
}

BOOL DestroyInstance ( DWORD InstanceIndex )
{
	if ( 0 == --dwInstCount )
	{
		hooking_void ( );
		dsp_void ( );
	}
	return TRUE;
}


/* library entry */
BOOL WINAPI DllMain ( HANDLE hinstDLL, DWORD dwReason, LPVOID lpvReserved )
{
	if ( ( DLL_PROCESS_ATTACH == dwReason ) && lpvReserved )
	{
	}
	return TRUE;
}
