/********************************************************************************
 *   dbgreader -- system debug output reader for Windows CE                     *
 *                                                                              * 
 *   Copyright (C) 2004-2007 dci (dev@cpp.in)                                   *
 *                                                                              *
 *   This program is free software; you can redistribute it and/or              *
 *   modify it under the terms of the GNU General Public License version 2      *
 *   as published by the Free Software Foundation.                              *
 *                                                                              *
 *   This program is distributed in the hope that it will be useful,            *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU          *
 *   General Public License version 2 for more details.                         *
 *                                                                              *
 *   You should have received a copy of the GNU General Public License          *
 *   version 2 along with this library; if not, write to the Free Software      *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  *
 ********************************************************************************
 */


/*
 static pool by dci.
   pool  maximum size: sizeof ( poolblock_t ) * min ( sizeof ( poolblock_t.prev ), min ( sizeof ( poolblock_t.next ), sizeof ( poolblock_t.upto ) ) );
   block maximum size: sizeof ( poolblock_t ) * sizeof ( poolblock_t.used );
*/

#include "staticpool.h"

/* memory block header struct. sizeof() must be /4 ! */
typedef struct
{
	unsigned short prev; /* number of sizeof(poolblock_t) from beginning of this block to beginning of next */
	unsigned short next; /* number of sizeof(poolblock_t) from beginning of prev block to beginning of this */
	unsigned short used; /* number of sizeof(poolblock_t), that block has allocated */
	unsigned short upto; /* number of sizeof(poolblock_t), that block was able to allocate */
}
poolblock_t;

/* number of elements of specified size, that fits(or partially overlaps) in memory block of specified size */
#define DCI_ITEMS_IN(tsize,bsize) ( ( tsize / bsize ) + ( ( tsize % bsize ) ? 1 : 0 ) )

/* static storage */
poolblock_t gb_pool [ 1 + DCI_ITEMS_IN ( DCI_POOL_BYTES, sizeof ( poolblock_t ) ) ]; 

/* funcs */
void dsp_init ( )
{
	/* init first block record */
	dsp_void ( );
	gb_pool->upto = ( unsigned short ) ( sizeof ( gb_pool ) / sizeof ( * gb_pool ) ) - 1;
}

void dsp_void ( )
{
	/* empty first block record */
	gb_pool->prev = 0;
	gb_pool->next = 0;
	gb_pool->used = 0;
	gb_pool->upto = 0;
}


void * dsp_alloc ( unsigned short rsize )
{
	poolblock_t * lb_cb = gb_pool;
	/* convert rsize to sizeof(poolblock_t)-based */
	rsize = DCI_ITEMS_IN ( rsize, sizeof ( poolblock_t ) );
	/* alloc */
	while ( 1 )
	{
		/* try to alloc from current block's wasted space */
		if ( ( lb_cb->upto - lb_cb->used ) >= ( rsize + 1 ) ) /* we'll fit in! */
		{
			poolblock_t * lb_new = lb_cb + 1 + lb_cb->used;
			/* set up next */
			if ( lb_cb->next )
			{
				poolblock_t * lb_next = lb_cb + lb_cb->next;
				lb_new->next = lb_next - lb_new;
				lb_next->prev = lb_new->next;
			}
			lb_cb->next = ( lb_new - lb_cb );
			/* set up prev */
			lb_new->prev = lb_cb->next;
			/* set up used */
			lb_new->used = rsize;
			/* set up upto */
			lb_new->upto = ( lb_cb->upto - lb_cb->used ) - 1;
			lb_cb->upto -= ( lb_new->upto + 1 );
			/* return ptr */
			return ( void * ) ( lb_new + 1 );
		}
		/* try to move ptr to next block */
		if ( lb_cb->next ) lb_cb += lb_cb->next;
		else return 0; /* no more blocks */
	}
}

void * dsp_realloc ( void * block, unsigned short newsize )
{
	poolblock_t * lb_cb = ( ( poolblock_t * ) block ) - 1;
	unsigned short blocks = DCI_ITEMS_IN ( newsize, sizeof ( poolblock_t ) );
	if ( blocks <= lb_cb->upto ) /* can fit in current block? */
	{
		lb_cb->used = blocks;
		return block;
	}
	else /* move */
	{
		void * ret = dsp_alloc ( newsize );
		if ( ret )
		{
			unsigned short i;
			for ( i = 0; i < ( lb_cb->used * sizeof ( poolblock_t ) ); ++i )
			{
				( ( unsigned char * ) ret ) [ i ] = ( ( unsigned char * ) block ) [ i ];
			}
			dsp_free ( block );
		}
		return ret;
	}
}

void dsp_free ( void * block )
{
	poolblock_t * lb_cb = ( ( poolblock_t * ) block ) - 1;
	/* free */
	lb_cb->used = 0;
	/* crosslink */
	if ( lb_cb->prev )
	{
		poolblock_t * lb_prev = lb_cb - lb_cb->prev;
		/* update prev's next(and next's prev, if it exists) */
		if ( lb_cb->next )
		{
			poolblock_t * lb_next = lb_cb + lb_cb->next;
			lb_prev->next = lb_next - lb_prev;
			lb_next->prev = lb_prev->next;
		}
		else lb_prev->next = 0;
		/* update prev's upto */
		lb_prev->upto += ( lb_cb->upto + 1 );
	}
}
